----------------------------------------------------------------------------------
-- Company: 
-- Engineer: 
-- 
-- Create Date:    22:01:44 04/15/2010 
-- Design Name: 
-- Module Name:    BoardMemSel - Behavioral 
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description: 
--
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

---- Uncomment the following library declaration if instantiating
---- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

entity BoardMemSel is
    Port ( cpuclk : in  STD_LOGIC;
           addr_64k : in  STD_LOGIC_VECTOR (23 downto 16);
           addr_8k : in  STD_LOGIC_VECTOR (15 downto 13);
			  isorig : in std_logic;
           validaddr : in  STD_LOGIC;
           cpurnw : in  STD_LOGIC;
           reset : in  STD_LOGIC;
           prgrom : in  STD_LOGIC;
           bootrom : in  STD_LOGIC;
           wprotect : in  STD_LOGIC_VECTOR (1 downto 0);
			  
			  nramcs : out std_logic_vector (1 downto 0);
			  nromcs : out std_logic;
			  nromwe : out std_logic
		);
end BoardMemSel;

architecture Behavioral of BoardMemSel is

		signal islow64k : std_logic;
		
begin

		Low64: process(addr_64k)
		begin
			if (addr_64k = "00000000") then
				islow64k <= '1';
			else
				islow64k <= '0';
			end if;
		end process Low64;
		
		-- ram0 is always selected on write as well, but on read only where not slowdetected
		-- each one has 512k
		MemSel: process(cpuclk, addr_64k, cpurnw, reset, validaddr, 
						prgrom, bootrom, addr_8k, wprotect,
						islow64k, isorig)
		begin
			if (reset = '0' or validaddr = '0' or cpuclk = '0') then
				nramcs <= "11";
				nromcs <= '1';
				nromwe <= '1';
			else
				-- bank 0 ram, 0-512k
				if (addr_64k(23 downto 19) = "00000")
					then
					if islow64k = '1' then
						if (cpurnw = '1' and (bootrom = '1' or isorig = '1')) then
							-- if read, but bootrom or isorig enabled, then not selected
							nramcs(0) <= '1';
						else
							-- write, or read and not bootrom and not isorig
							-- check write protect in low 64k
							case wprotect is
								when "00" =>
									nramcs(0) <= '0';		-- no write protect
								when "01" =>
									if (addr_8k(15) = '1'
										and (addr_8k(14) = '1'
											or addr_8k(13) = '1')) then
										nramcs(0) <= not(cpurnw);		-- write protected $a000-$ffff
									else
										nramcs(0) <= '0';	-- ok
									end if;
								when "10" =>
									if (addr_8k(15) = '1'
										and addr_8k(14) = '1') then
										nramcs(0) <= not(cpurnw);		-- write protected $c000-$ffff
									else
										nramcs(0) <= '0';	-- ok
									end if;
								when "11" =>
									if (addr_8k(15) = '1'
										and addr_8k(14) = '1'
										and addr_8k(13) = '1') then
										nramcs(0) <= not(cpurnw);		-- write protected $e000-$ffff
									else
										nramcs(0) <= '0';	-- ok
									end if;
								when others =>
									nramcs(0) <= '1';						-- no access, should not happen
							end case;
						end if;  -- read/bootrom/orig
					else	-- if islow64k
						-- 64k-512k
						nramcs(0) <= '0';
					end if;
				elsif (addr_64k(23 downto 19) = "11110") then
					-- mirrored at $f0xxxx-$f7xxxx
					nramcs(0) <= '0';
				else
					nramcs(0) <= '1';
				end if;
			
				-- bank 1 ram, 512k-1M
				if (addr_64k(23 downto 19) = "00001")
					then
					-- $08xxxx-$0fxxxx
					nramcs(1) <= '0';
				else
					nramcs(1) <= '1';
				end if;

				-- uppermost 512k, boot rom
				if (addr_64k(23 downto 19) = "11111")
					then
					-- $f8xxxx-$ffxxxx
					-- uppermost 512k, write only if allowed by prgrom
					nromcs <= not(cpurnw or prgrom);
					nromwe <= cpurnw;
				elsif (islow64k = '1' 
						and cpurnw = '1'
						and bootrom = '1' 
						and isorig = '0'
						) then 
						-- low 64k, bootrom, read only, not I/O and not video
						nromcs <= '0';
						nromwe <= '1';
				else
					nromcs <= '1';
					nromwe <= '1';
				end if;
			end if;

		end process MemSel;


end Behavioral;

