----------------------------------------------------------------------------------
-- Company: 
-- Engineer: 
-- 
-- Create Date:    23:06:54 02/03/2010 
-- Design Name: 
-- Module Name:    PET816 - Behavioral 
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description: 
--
-- Dependencies: 
--
-- Revision: 
-- Revision 1.0A - First revision
-- Additional Comments: ----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;
-- use IEEE.numeric_std.all;
-- use ieee.std_logic_unsigned.all;

---- Uncomment the following library declaration if instantiating
---- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

entity PET816 is
	Port( 			
			fastclkin : in STD_LOGIC;		-- 50MHz in
			
			-- PET interface

			phi0 : in STD_LOGIC;				-- PET Phi0 in
			phi1 : out STD_LOGIC;			-- PET Phi1 out
			phi2 : out STD_LOGIC;			-- PET Phi2 out
			rdy : in STD_LOGIC;
			reset : in STD_LOGIC;
			
			rnw : out STD_LOGIC;
			
			-- 65816 interface

			-- if you integrate the '573 high address byte latch into the CPLD, comment the next line
			-- and uncomment the following line
			cpuaddrin_64k : in STD_LOGIC_VECTOR (23 downto 16);
			--cpuaddrout_64k : out STD_LOGIC_VECTOR (18 downto 16);
			
			cpuaddrin_2k : in STD_LOGIC_VECTOR (15 downto 11);
			cpuaddr_ismyio : in STD_LOGIC;		-- when 0 then addr bits 4 to 11 are all 1
			cpuaddrin_16 : in STD_LOGIC_VECTOR (3 downto 0);
			cpudata : inout STD_LOGIC_VECTOR (7 downto 0);
			
			cpuvpa : in STD_LOGIC;
			cpuvda : in STD_LOGIC;
			--cpuvp : in STD_LOGIC;
			cpurnw : in STD_LOGIC;
			cpuwnr : out STD_LOGIC;			-- inverted CPU r/-w
			
			cpuclk : out STD_LOGIC;
			
			-- control
			
			slowaddrlatch : out STD_LOGIC;			-- when 0, pass CPU address bus to PET
			nslowbusclr	:out std_logic;			-- when 0, addr PET addr is set to 0
			nslowdataenout : out STD_LOGIC;			-- when 0, pass CPU data bus to PET
			latchen: out STD_LOGIC;					-- '573 latch control (1=passthrough, 0=latch)
			
			-- config
			bootromin: in STD_LOGIC;				-- when 1 then jumper is set to enable boot rom
			
			-- fast memory interface (2 RAM modules)
			
			nramcs : out STD_LOGIC_VECTOR (1 downto 0);
			
			-- bootrom select, in-system-programming
			nromcs : out STD_LOGIC;
			nromwe : out STD_LOGIC;
						
			-- diagnostic output signal
			--diag : out STD_LOGIC
			
			-- auxiliary config input
			auxconf : in std_logic
	);
			
end PET816;

architecture Behavioral of PET816 is

		-- internal state signals
		signal useslowclk: std_logic := '1';				-- when 1 then the slow bus clock should be passed through to the cpu
		signal hidebogus: std_logic := '0';					-- when 1 then bogus accesses are hidden from the slow bus
		signal slow64k: std_logic := '1';					-- when 1 low 64k are mapped to PET (otherwise I/O, video only)
																		--	writes are passed through to fast RAM (with slow access)
		signal bootrom: std_logic := '0';					-- when 1, map expansion ROM to low 64k; initialied to jumper
		signal prgrom: std_logic := '0';						-- when 1, allow writes to ROM (in system programming)
		signal wprotect: std_logic_vector (1 downto 0);	-- when 1, write protect fast RAM map
		signal fastvread: std_logic := '0';					-- when 1, read video memory from fast RAM
		
		-- master timing
		signal masterclk		:std_logic := '0';					-- fast clock divided by clkcount
		signal fastclkby2		:std_logic := '0';					-- fastclkin divided by 2
		signal clklatch		:std_logic_vector (4 downto 0);
		signal clklatchx		:std_logic_vector (4 downto 0);	-- for slowclk
		signal slowclk 		:std_logic;
		
		signal extclkdel: std_logic := '0';			-- delayed phi0; a slow bus access starts at the falling edge
		
		-- state engine
		signal slowaccess: std_logic := '0';
		signal clkenable: std_logic := '0';
		signal startfast: std_logic := '0';
		signal slowdetected	:std_logic;		
		
		signal cpuclkint: std_logic := '0';
		
		signal validaddr: std_logic := '0';					-- vpa/vda/vp analysis
		
		signal isio: std_logic := '0';			-- 1 when the I/O area is selected (valid together with islow64k)
		signal isvideo: std_logic := '0';		-- 1 when the video area is selected (valid together with islow64k)
		signal islow64k: std_logic := '0';		-- 1 when the lowest 64k are selected
		signal isorig: std_logic := '0';			-- 1 when the original system is being addressed
	
		signal busaccess	:std_logic := '0';		-- 1 when bus is accessed via $Exxxxx

		-- if you integrate the '573 latch into the CPLD, use these as replacement for the input signals
		--	signal cpuaddrin_64k : STD_LOGIC_VECTOR (23 downto 16);

		component ClockDivider
			Port ( 
			  fastclkin : in  STD_LOGIC;
           fastclkby2 : out  STD_LOGIC;
           fastclkbydiv : out  STD_LOGIC;
			  divenable : in STD_LOGIC;
           clklatch : in  STD_LOGIC_VECTOR (4 downto 0);
			  reset : in STD_LOGIC
			);
		end component;

		component ClockShift
			Port (
				phi : in std_logic;
				fastclk : in std_logic;
				fastclkby2 : in std_logic;
				shortdelay : in std_logic;
				phidelayed : out std_logic;
				pulser : out std_logic;
				pulsew : out std_logic
			);
		end component;
		
		component ControlRegister
		  Port ( 
			  data : inout  STD_LOGIC_VECTOR (7 downto 0);
           clk : in  STD_LOGIC;
			  sel : in STD_LOGIC;
			  addr : in STD_LOGIC;
			  rnw : in STD_LOGIC;
           reset : in  STD_LOGIC;
           bootromin : in  STD_LOGIC;
			  auxconf : in std_logic;
			  
			  -- control register
           wprotect : out  STD_LOGIC_VECTOR (1 downto 0);
           hidebogus : out  STD_LOGIC;
           bootrom : out  STD_LOGIC;
           prgrom : out  STD_LOGIC;
			  slow64k : out STD_LOGIC;
			  slowclk : out STD_LOGIC;
			  fastvread : out STD_LOGIC;
			  
			  -- clock control
           clklatch : out  STD_LOGIC_VECTOR (4 downto 0);
			  fastmode : out std_logic
		  );
		end component;

		signal ctrlsel : STD_LOGIC;

		component BoardMemSel
		  Port ( cpuclk : in  STD_LOGIC;
           addr_64k : in  STD_LOGIC_VECTOR (23 downto 16);
           addr_8k : in  STD_LOGIC_VECTOR (15 downto 13);
			  isorig : in std_logic;
           validaddr : in  STD_LOGIC;
           cpurnw : in  STD_LOGIC;
           reset : in  STD_LOGIC;
           prgrom : in  STD_LOGIC;
           bootrom : in  STD_LOGIC;
           wprotect : in  STD_LOGIC_VECTOR (1 downto 0);
			  
			  nramcs : out std_logic_vector(1 downto 0);
			  nromcs : out std_logic;
			  nromwe : out std_logic
			);
		end component;

		signal nromcsint : std_logic;
		
		component ClockSync 
			Port ( 
			  slowdetectedout : out std_logic;
			  slowaccessout : out  STD_LOGIC;
			  cpuclk : in std_logic;
           startfast : in  STD_LOGIC;
           extclkdel : in  STD_LOGIC;
           extclk : in  STD_LOGIC;
			  rdy : in std_logic;
			  reset : in std_logic;
			  isslow : in std_logic
		);
		end component;

		component SystemMap is
			Port ( 
			  cpuaddrin_64k : in  STD_LOGIC_VECTOR (23 downto 16);
           cpuaddrin_2k : in  STD_LOGIC_VECTOR (15 downto 11);
			  cpuaddrin_ismyio : in STD_LOGIC;
           cpuaddrin_16 : in  STD_LOGIC_VECTOR (3 downto 0);
			  islow64k : in STD_LOGIC;
           isio : out  STD_LOGIC;
           isvideo : out  STD_LOGIC;
           busaccess : out  STD_LOGIC;
           romaccess : out  STD_LOGIC;
			  ctrlsel : out STD_LOGIC
			);
		end component;

		signal isslow : std_logic;
		signal romaccess : std_logic;
		
		signal startfastr 	:std_logic;			-- ends slow read cycle
		signal startfastw		:std_logic;			-- ends slow write cycle (later, to obey data hold time)
		
		signal fastmode		:std_logic;			-- to enable 2MHz host system
		
		signal nslowdataen	:std_logic;
		
		signal rnwint			:std_logic;			-- not-clocked rnw
		
		signal origdetected	:std_logic;
		
begin
		-----------------------------------------------------------------------------------------------
		-- system specific part

		DoMap: process(cpuaddrin_2k, cpuaddrin_64k, slow64k, bootrom)
		begin
			-- lowest 64k address area, gets special handling
			if (cpuaddrin_64k = "00000000") then
				islow64k <= '1';
			else
				islow64k <= '0';
			end if;
		end process;

		MyMap: SystemMap port map (
				cpuaddrin_64k,
				cpuaddrin_2k,
				cpuaddr_ismyio,
				cpuaddrin_16,
				islow64k,
				isio,
				isvideo,
				busaccess,
				romaccess,
				ctrlsel
				);
		
		-- enables bus interface
		isorig <= (isvideo and not(fastvread and cpurnw)) 
			  or (isio and not(ctrlsel))
			  or busaccess
			  or (islow64k and slow64k and not(bootrom));

		-----------------------------------------------------------------------------------------------
		-- CPU816 uppermost address byte latched from the data bus
		
--		HighAddr: process(cpuclkint, cpudata)
--		begin
--			if (cpuclkint = '0') then
--				cpuaddrin_64k <= cpudata;
--			end if;
--		end process;
--		
--		cpuaddrout_64k <= cpuaddrin_64k(18 downto 16);
		
		-----------------------------------------------------------------------------------------------
		-- master timing 
	
		-- that leaves us at least a fastclk phi1 phase to recover from slow accesses
		DoClkEnable: process(slowdetected, reset, useslowclk, startfast)
		begin
			if (reset = '0') then
				clkenable <= '1';
			elsif (slowdetected = '1') then
				clkenable <= '0';
			elsif (rising_edge(startfast)) then
				clkenable <= '1';
			end if;
		end process;

		DoClkLatchX: process (clklatch, useslowclk)
		begin
			if (useslowclk = '1') then
				clklatchx <= "00100";
			else
				clklatchx <= clklatch;
			end if;
		end process;
		
		Mytiming: ClockDivider port map (
				fastclkin,
				fastclkby2,
				masterclk,
				clkenable,
				clklatchx,
				reset
				);
			
		slowclk <= phi0;
		phi1 <= not(slowclk);
		phi2 <= slowclk;
		
		cpuwnr <= not(cpurnw);
		
		-----------------------------------------------------------------------------------------------
		-- bus timing generation
		-- pass phi1 through shift register by fastclock (50MHz)
		-- @50MHz -> @25MHz fastclkb2, each step is 40ns
		
		MyClock: ClockShift port map (
				slowclk,
				fastclkin,
				fastclkby2,
				fastmode,
				extclkdel,
				startfastr,
				startfastw
				);
				
		-----------------------------------------------------------------------------------------------
		-- cpu bus timing

		-- synchronize all signals with the falling edge of extclkdel
		-- slowaddrlatch: low-high transition transfers the cpu addr to the system addr bus
		-- nslowbusclr: if low, set system addr bus to zero
		-- nslowdataenout: if low, enable system data bus; 
		--                 note: must be disabled during phi1, to get CPU segment byte of address
		-- rnw: system r/-w line
		
		origdetected <= slowdetected and isorig;
		
		BusTiming: process(reset, extclkdel, cpurnw, origdetected, startfast)
		begin
			if (reset = '0') then 
				slowaddrlatch <= '1';
				nslowbusclr <= '0';
--				nslowdataen <= '1';
				rnwint <= '0';
			else
				if (origdetected = '1') then
					nslowbusclr <= '1';
				elsif (falling_edge(extclkdel)) then
					nslowbusclr <= origdetected;	-- goes low when NO slow access
				end if;
				
				if (startfast = '1') then
					slowaddrlatch <= '0';
					nslowdataen <= '1';
				elsif (falling_edge(extclkdel)) then
					slowaddrlatch <= origdetected; -- goes high when slow access starts
					nslowdataen <= not(origdetected);
				end if;

				if (falling_edge(extclkdel)) then
					rnwint <= cpurnw or not(origdetected);
				end if;
			end if;
		end process;
		
		rnw <= rnwint;
		
		-- delay data enable, so 
		-- a) '573 can finish loading the high order address byte before the slow bus driver cuts in 
		--    (slowdetected is going active on fast phi2 going high) and
		-- b) increase the hold time of the data bus (for writes)
		DoSlowData: process(fastclkin, nslowdataen, reset)
		begin
			if (reset = '0') then
				nslowdataenout <= '1';
			elsif (rising_edge(fastclkin)) then
				nslowdataenout <= nslowdataen;
			end if;
		end process;

		-----------------------------------------------------------------------------------------------
		
		-- CPU clock, syntesised
		cpuclkint <= 
				(
					(masterclk)		-- fast access
					or (slowdetected)
					or (slowaccess) -- extends beyond slowdetected! It is strangely important that reads are extended too!
				);
				
		cpuclk <= cpuclkint;
		latchen <= not(cpuclkint);

		--diag <= fakeclk;

		validaddr <= cpuvpa or cpuvda;
		
		-----------------------------------------------------------------------------------------------
		-- synchronize the fast and the slow clock
		
		isslow <= useslowclk or 
					( (validaddr or not(hidebogus)) and ( romaccess or isorig ));
					
		startfast <= not(slowdetected) 
					and ((rnwint and startfastr) 
						or (not(rnwint) and startfastw));
		
		MyClockSync: ClockSync port map (
			  slowdetected,
			  slowaccess,
			  cpuclkint,
           startfast,
           extclkdel,
           slowclk,
			  rdy,
			  reset,
			  isslow
		);
		
		-----------------------------------------------------------------------------------------------
		-- CPU access to config registers
		-- 2 byte address space

		
		MyControl : ControlRegister port map (
			cpudata,
			cpuclkint,
			ctrlsel,
			cpuaddrin_16(0),
			cpurnw,
			reset,
			bootromin,
			auxconf,
			wprotect,
			hidebogus,
			bootrom,
			prgrom,
			slow64k,
			useslowclk,
			fastvread,
			clklatch,
			fastmode
		);
		
		-----------------------------------------------------------------------------------------------
		-- memory select lines

		--slowdetected <= slowaccess or fakeclk;
		
		memsel : BoardMemSel port map (
			cpuclkint,
			cpuaddrin_64k,
			cpuaddrin_2k(15 downto 13),
			isorig,
			validaddr,
			cpurnw,
			reset,
			prgrom,
			bootrom,
			wprotect,
			
			nramcs,
			nromcsint,
			nromwe
		);

		nromcs <= nromcsint or not(slowclk and slowaccess);
		
end Behavioral;

